#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ipc.h> 
#include <sys/shm.h> 
#include <sys/mman.h>
#include <sys/time.h> 
#include "address_map_arm_brl4.h"
#include <math.h>

//===Offsets from Base====//
#define FFT_ADDR          0x00000060
#define FFT_DATA          0x00000030
#define FFT_VALID         0x00000010


#define NUM_ENEMIES			25
#define SPAWN_TIME			1000
#define ENEMY_RAD			4
#define PLAYER_RAD			12
#define PLAYER_MOVE			8
#define MAXFFT 				10
#define NUM_STARS			200
#define PLAY_HEALTH			3
#define PLAY_COLOR			0X49
#define ENEMY_COL			0X88

// Player Struct
typedef struct player {
	float xpos;
	float ypos;
	float prevx;
	float prevy;
	int health;
} player;

// Enemy Struct
typedef struct enemy {
	float xpos;
	float ypos;
	float prevx;
	float prevy;
	float xvel;
	float yvel;
	int valid;
	int radius;
} enemy;

typedef struct star {
	float xpos;
	float ypos;
	float prevx;
	float xvel;
} star;

/* function prototypes */
void VGA_text (int, int, char *);
void VGA_text_clear();
void VGA_box (int, int, int, int, short);
void VGA_line(int, int, int, int, short);
void VGA_disc (int, int, int, short);

void drawPlayer(player*, short);
void initPlayer(player*);
void drawEnemy(enemy*, short);
void updateStars();
void updateEnemies();
void drawEnemies(short);
void spawnEnemy();
void detectCollision();
void melspec(float[], int, float[], int, float);
double window(int, int);
void dct(float[], float[], int);


// the light weight bus base
void *h2p_lw_virtual_base;
volatile unsigned int * h2p_lw_sw_addr = NULL ;

// pixel buffer
volatile unsigned int * vga_pixel_ptr = NULL ;
void *vga_pixel_virtual_base;

// character buffer
volatile unsigned int * vga_char_ptr = NULL ;
void *vga_char_virtual_base;

// /dev/mem file id
int fd;

player thePlayer;
enemy enemies[NUM_ENEMIES];
star stars[NUM_STARS];

// measure time
struct timeval t1, t2, t3, t4;
double elapsedTime, lastSpawn, loopTime, elapsed;

float mels[12];
float mfcc1[12];
float fs[] = {50,205.61,393.51,620.4,894.36,1225.17,1624.61,2106.93,2689.32,3392.55,4241.69,5267.01,6505.06,8000};
float ffts[128];
float bigfft[128*MAXFFT];
int mode;


int main(void) {

	// === get FPGA addresses ==================
    // Open /dev/mem
	if( ( fd = open( "/dev/mem", ( O_RDWR | O_SYNC ) ) ) == -1 ) {
		printf( "ERROR: could not open \"/dev/mem\"...\n" );
		return( 1 );
	}
    
    // get virtual addr that maps to physical
	h2p_lw_virtual_base = mmap( NULL, HW_REGS_SPAN, ( PROT_READ | PROT_WRITE ), MAP_SHARED, fd, HW_REGS_BASE );	
	if( h2p_lw_virtual_base == MAP_FAILED ) {
		printf( "ERROR: mmap1() failed...\n" );
		close( fd );
		return(1);
	}

	// === get VGA char addr =====================
	// get virtual addr that maps to physical
	vga_char_virtual_base = mmap( NULL, FPGA_CHAR_SPAN, ( 	PROT_READ | PROT_WRITE ), MAP_SHARED, fd, FPGA_CHAR_BASE );	
	if( vga_char_virtual_base == MAP_FAILED ) {
		printf( "ERROR: mmap2() failed...\n" );
		close( fd );
		return(1);
	}
    
    // Get the address that maps to the FPGA LED control 
	vga_char_ptr =(unsigned int *)(vga_char_virtual_base);

	// === get VGA pixel addr ====================
	// get virtual addr that maps to physical
	vga_pixel_virtual_base = mmap( NULL, FPGA_ONCHIP_SPAN, ( 	PROT_READ | PROT_WRITE ), MAP_SHARED, fd, 			FPGA_ONCHIP_BASE);	
	if( vga_pixel_virtual_base == MAP_FAILED ) {
		printf( "ERROR: mmap3() failed...\n" );
		close( fd );
		return(1);
	}
    
    // Get the address that maps to the FPGA pixel buffer
	vga_pixel_ptr =(unsigned int *)(vga_pixel_virtual_base);


	/* create a message to be displayed on the VGA and LCD displays */
	char text_top_row[40] = "Cool Game\0";
	char text_bottom_row[40] = "Press Enter to start\0";
	char num_string[20], time_string[40], health_string[40], score_string[40];

	// clear the screen
	VGA_box (0, 0, 639, 479, 0x00);
	mode = 0;
	int spawn = SPAWN_TIME;

	int j;
	for (j=0; j < NUM_STARS; j++) {
		stars[j].xpos = rand() % 640;
		stars[j].ypos = rand() % 480;
		stars[j].prevx = stars[j].xpos;
		if (j%3 == 0) stars[j].xvel = -.01;
		else if (j%3 == 1) stars[j].xvel = -.04;
		else stars[j].xvel = -.07;
	}
	updateStars();
	// clear the text
	VGA_text_clear();
	VGA_text (35, 18, text_top_row);
	VGA_text (30, 22, text_bottom_row);



	char input = getchar();	// Get first character of inputted string
	if (input) {  //TODO: check this works
		if (input == 'H' || input == 'h') {
					spawn = 300;
					mode = 1;
		}
		printf("Starting Game \n");
	}	

	VGA_text_clear();
	VGA_box (0, 0, 639, 479, 0x00);

	initPlayer(&thePlayer);

	int i, num, max;
	int state = 0;
	int counter;
	int threshold = 77;
	

	while(1) {

		if (thePlayer.health == 0) {
			// clear the screen
			VGA_box (0, 0, 639, 479, 0x00);
			// clear the text
			VGA_text_clear();
			VGA_text (36, 18, "You Lose :(");
		 	sprintf(score_string, "You lasted for %.0f seconds", elapsedTime/1000);
		 	VGA_text(30, 20, score_string);
			VGA_text (31, 22, "Press Enter to restart");
			char input = getchar();	// Get first character of inputted string
			if (input) {  
				if (input == 'H' || input == 'h') {
					spawn = 300;
					mode = 1;
				}
				int i;
				for (i = 0; i < NUM_ENEMIES; i++) {
					enemies[i].valid = 0;
					drawEnemy(&enemies[i], 0x00);
				}
				VGA_text_clear();
				elapsedTime = 0;
				loopTime = 0;
				lastSpawn = 0;
				elapsed = 0;
				initPlayer(&thePlayer);
				printf("Starting Game\n");
			}	
		}

		gettimeofday(&t3, NULL);

		if (elapsedTime - lastSpawn > spawn) {
			spawnEnemy();
			lastSpawn = elapsedTime;
		}
		updateStars();
		updateEnemies();
		drawPlayer(&thePlayer, PLAY_COLOR);
		detectCollision();
		

		while (loopTime < 16) {
			if (*(unsigned int*)(h2p_lw_virtual_base + FFT_VALID)) {
				max = 0;
				for (i = 0; i <= 127; i++) {
					*(unsigned int*)(h2p_lw_virtual_base + FFT_ADDR) = 127 - i;
					usleep(1);
					num = *(unsigned int*)(h2p_lw_virtual_base + FFT_DATA);
					ffts[i] = num;
					if (num > max) {
						max = num;
					}
				}
				if (state == 0) {
					if (max > 900000) {
						gettimeofday(&t1, NULL);
						for (i=0;i<128;i++) {
							bigfft[i] = ffts[i];
						}
						counter = 0;
						state = 1;
					}
				}
				else if (state == 1) { //detecting start of word
					if (max > 900000) {
						gettimeofday(&t2, NULL);
						elapsed = (t2.tv_sec - t1.tv_sec) * 1000.0;      // sec to ms
						elapsed += (t2.tv_usec - t1.tv_usec) / 1000.0;   // us to ms
						if (elapsed > 100) { //probably just noise
							gettimeofday(&t1, NULL);
							for (i=0;i<128;i++) {
								bigfft[i] = ffts[i];
							}
							counter = 0;
						}
						else { //start of a word
							counter++;
						 	gettimeofday(&t1, NULL);
						 	for (i=0;i<128;i++) {
								bigfft[128*counter + i] = ffts[i];
							}
							state = 2;
						}
					}
				}
				else if (state == 2) { //stay here during word
					gettimeofday(&t2, NULL);
					elapsed = (t2.tv_sec - t1.tv_sec) * 1000.0;      // sec to ms
					elapsed += (t2.tv_usec - t1.tv_usec) / 1000.0;   // us to ms
					counter++;
					for (i=0;i<128;i++) {
						bigfft[128*counter + i] = ffts[i];
					}
					if (elapsed > 200 || counter >= MAXFFT-1) { //end of word
						melspec(bigfft, 128*(counter+1), mels, 12, 48000);
						dct(mels, mfcc1, 12);
						if (mfcc1[0] - mfcc1[1] > threshold) {
							thePlayer.prevy = thePlayer.ypos;
							drawPlayer(&thePlayer, 0x00);
							thePlayer.ypos -= PLAYER_MOVE;
							if (thePlayer.ypos < PLAYER_RAD)  thePlayer.ypos = PLAYER_RAD;
							drawPlayer(&thePlayer, PLAY_COLOR);
						}
						else {
							thePlayer.prevy = thePlayer.ypos;
							drawPlayer(&thePlayer, 0x00);
							thePlayer.ypos += PLAYER_MOVE;
							if (thePlayer.ypos > (479 - PLAYER_RAD) ) thePlayer.ypos = (479 - PLAYER_RAD);
							drawPlayer(&thePlayer, PLAY_COLOR);							
						}
						gettimeofday(&t1, NULL);
						for (i=0;i<128;i++) {
							bigfft[i] = ffts[i];
						}
						counter = 0;
						state = 1;
					}
					else if (max > 900000) { //still in word
					 	gettimeofday(&t1, NULL);
					}
				}
			}

			// stop timer
			 gettimeofday(&t4, NULL);
			 
			 loopTime = (t4.tv_sec - t3.tv_sec) * 1000.0;      // sec to ms
			 loopTime += (t4.tv_usec - t3.tv_usec) / 1000.0;   // us to ms
			
		}
		gettimeofday(&t4, NULL);
		elapsedTime += (t4.tv_sec - t3.tv_sec) * 1000.0;      // sec to ms
		elapsedTime += (t4.tv_usec - t3.tv_usec) / 1000.0;   // us to ms
		loopTime = 0;
		 sprintf(time_string, "T=%3.0fmS   ", elapsedTime);
		 sprintf(health_string, "Health=%d   ", thePlayer.health);
		 VGA_text (1, 1, time_string);
		 VGA_text (1, 2, health_string);

	}
}

// pixel macro
#define VGA_PIXEL(x,y,color) do{\
	char  *pixel_ptr ;\
	pixel_ptr = (char *)vga_pixel_ptr + ((y)<<10) + (x) ;\
	*(char *)pixel_ptr = (color);\
} while(0)

// =============================================
// === Draw a line
// =============================================
//plot a line 
//at x1,y1 to x2,y2 with color 
//Code is from David Rodgers,
//"Procedural Elements of Computer Graphics",1985
void VGA_line(int x1, int y1, int x2, int y2, short c) {
	int e;
	signed int dx,dy,j, temp;
	signed int s1,s2, xchange;
     signed int x,y;
	char *pixel_ptr ;
	
	/* check and fix line coordinates to be valid */
	if (x1>639) x1 = 639;
	if (y1>479) y1 = 479;
	if (x2>639) x2 = 639;
	if (y2>479) y2 = 479;
	if (x1<0) x1 = 0;
	if (y1<0) y1 = 0;
	if (x2<0) x2 = 0;
	if (y2<0) y2 = 0;
        
	x = x1;
	y = y1;
	
	//take absolute value
	if (x2 < x1) {
		dx = x1 - x2;
		s1 = -1;
	}

	else if (x2 == x1) {
		dx = 0;
		s1 = 0;
	}

	else {
		dx = x2 - x1;
		s1 = 1;
	}

	if (y2 < y1) {
		dy = y1 - y2;
		s2 = -1;
	}

	else if (y2 == y1) {
		dy = 0;
		s2 = 0;
	}

	else {
		dy = y2 - y1;
		s2 = 1;
	}

	xchange = 0;   

	if (dy>dx) {
		temp = dx;
		dx = dy;
		dy = temp;
		xchange = 1;
	} 

	e = ((int)dy<<1) - dx;  
	 
	for (j=0; j<=dx; j++) {
		//video_pt(x,y,c); //640x480
		pixel_ptr = (char *)vga_pixel_ptr + (y<<10)+ x; 
		// set pixel color
		*(char *)pixel_ptr = c;	
		 
		if (e>=0) {
			if (xchange==1) x = x + s1;
			else y = y + s2;
			e = e - ((int)dx<<1);
		}

		if (xchange==1) y = y + s2;
		else x = x + s1;

		e = e + ((int)dy<<1);
	}
}

/****************************************************************************************
 * Draw a filled rectangle on the VGA monitor 
****************************************************************************************/
#define SWAP(X,Y) do{int temp=X; X=Y; Y=temp;}while(0) 

void VGA_box(int x1, int y1, int x2, int y2, short pixel_color)
{
	char  *pixel_ptr ; 
	int row, col;

	/* check and fix box coordinates to be valid */
	if (x1>639) x1 = 639;
	if (y1>479) y1 = 479;
	if (x2>639) x2 = 639;
	if (y2>479) y2 = 479;
	if (x1<0) x1 = 0;
	if (y1<0) y1 = 0;
	if (x2<0) x2 = 0;
	if (y2<0) y2 = 0;
	if (x1>x2) SWAP(x1,x2);
	if (y1>y2) SWAP(y1,y2);
	for (row = y1; row <= y2; row++)
		for (col = x1; col <= x2; ++col)
		{
			//640x480
			pixel_ptr = (char *)vga_pixel_ptr + (row<<10)    + col ;
			// set pixel color
			*(char *)pixel_ptr = pixel_color;		
		}
}

/****************************************************************************************
 * Subroutine to send a string of text to the VGA monitor 
****************************************************************************************/
void VGA_text(int x, int y, char * text_ptr)
{
  	volatile char * character_buffer = (char *) vga_char_ptr ;	// VGA character buffer
	int offset;
	/* assume that the text string fits on one line */
	offset = (y << 7) + x;
	while ( *(text_ptr) )
	{
		// write to the character buffer
		*(character_buffer + offset) = *(text_ptr);	
		++text_ptr;
		++offset;
	}
}

/****************************************************************************************
 * Subroutine to clear text to the VGA monitor 
****************************************************************************************/
void VGA_text_clear()
{
  	volatile char * character_buffer = (char *) vga_char_ptr ;	// VGA character buffer
	int offset, x, y;
	for (x=0; x<70; x++){
		for (y=0; y<40; y++){
	/* assume that the text string fits on one line */
			offset = (y << 7) + x;
			// write to the character buffer
			*(character_buffer + offset) = ' ';		
		}
	}
}

/****************************************************************************************
 * Draw a filled circle on the VGA monitor 
****************************************************************************************/

void VGA_disc(int x, int y, int r, short pixel_color)
{
	char  *pixel_ptr ; 
	int row, col, rsqr, xc, yc;
	
	rsqr = r*r;
	
	for (yc = -r; yc <= r; yc++)
		for (xc = -r; xc <= r; xc++)
		{
			col = xc;
			row = yc;
			// add the r to make the edge smoother
			if(col*col+row*row <= rsqr+r){
				col += x; // add the center point
				row += y; // add the center point
				//check for valid 640x480
				if (col>639) col = 639;
				if (row>479) row = 479;
				if (col<0) col = 0;
				if (row<0) row = 0;
				pixel_ptr = (char *)vga_pixel_ptr + (row<<10) + col ;
				// set pixel color
				*(char *)pixel_ptr = pixel_color;
			}
					
		}
}

void drawPlayer(player* player, short pixel_color) {
	VGA_box((int) player->xpos - PLAYER_RAD, (int) player->ypos - PLAYER_RAD, (int) player->xpos + PLAYER_RAD, (int) player->ypos + PLAYER_RAD, pixel_color);
}

void initPlayer(player* player) {
	player->xpos = 1 + PLAYER_RAD;
	player->ypos = 240;
	player->prevx = player->xpos;
	player->prevy = 240;
	player->health = PLAY_HEALTH;
	drawPlayer(player, PLAY_COLOR);
}

void drawEnemy(enemy* enemy, short pixel_color) {
	VGA_disc((int) enemy->xpos, (int) enemy->ypos, enemy->radius, pixel_color);
}

void updateEnemies() {
	int i;
	for (i = 0; i < NUM_ENEMIES; i++) {
		if (enemies[i].valid) {
			enemies[i].prevx = enemies[i].xpos;
			enemies[i].prevy = enemies[i].ypos;
			drawEnemy(&enemies[i], 0x00);
			enemies[i].xpos += enemies[i].xvel;
			enemies[i].ypos += enemies[i].yvel;
			if (enemies[i].xpos < 2 || enemies[i].ypos > 479 || enemies[i].ypos < 2) { //TODO: check these conditions with ENEMY_RAD
				enemies[i].valid = 0;
				enemies[i].xvel = 0;
				enemies[i].yvel = 0;
			} else {
				drawEnemy(&enemies[i], ENEMY_COL);
			}
		}
	}
}

void updateStars() {
	int i;
	for (i = 0; i < NUM_STARS; i++) {
		stars[i].prevx = stars[i].xpos;
		VGA_PIXEL((int)stars[i].xpos, (int)stars[i].ypos, 0x00);
		stars[i].xpos += stars[i].xvel;
		if (stars[i].xpos < 0) stars[i].xpos = 639;
		VGA_PIXEL((int)stars[i].xpos, (int)stars[i].ypos, 0xFF);
	}
}

void spawnEnemy() {
	int i;
	for (i = 0; i < NUM_ENEMIES; i++) {
		if (!enemies[i].valid) {

			enemies[i].xpos = 630;
			// range for rand: rand() % (max_number + 1 - minimum_number) + minimum_number
			enemies[i].ypos = rand() % (460 + 1 - 20) + 20;
			enemies[i].prevx = enemies[i].xpos;
			enemies[i].prevy = enemies[i].ypos;
			enemies[i].xvel = -((rand() % 1800) + 600)/1000.0;
			enemies[i].yvel = ((rand() % 401) - 200)/1000.0;
			enemies[i].valid = 1;
			if (mode == 1) enemies[i].radius = rand() % (15 + 1 -3) +3;
			else enemies[i].radius = 4;
			break;
		}
	}
}

void detectCollision() {
	int i;
	for (i = 0; i < NUM_ENEMIES; i++) {
		if (enemies[i].valid) {
			if (enemies[i].xpos < thePlayer.xpos + PLAYER_RAD + enemies[i].radius) {
				if ((enemies[i].ypos - thePlayer.ypos <= enemies[i].radius + PLAYER_RAD) && (enemies[i].ypos - thePlayer.ypos) >= -(enemies[i].radius + PLAYER_RAD)) {
					drawEnemy(&enemies[i], 0x00);
					enemies[i].valid = 0;
					thePlayer.health -= 1;
				}
			}
		}
	}

}

//convert n point fft spectrum to m point mel spectrum using triangular windows
void melspec(float in[], int n, float out[], int m, float fs)
{
    int i, j;
    float deltaf = fs / n;
    //loop through mel windows
    for (i = 0; i < m; i++) {
        out[i] = 0;
        //accumulate fft powers
        for (j = 0; j < n; j++) {
            out[i] += in[j]*window(j*deltaf, i+1);
        }
        out[i] = log10(out[i]);
    }
}

// power at frequency f in triangular window i
double window(int f, int i)
{
    if (f < fs[i-1] || f >= fs[i+1]) return 0;
    if (f < fs[i]) return (f-fs[i-1])/(fs[i]-fs[i-1]);
    else           return (f-fs[i+1])/(fs[i]-fs[i+1]);
}

//discrete cosine transform
void dct(float in[], float out[], int n)
{
    int i,j;
    double pn = M_PI / n;
    for (i = 0; i < n; i++) {
        out[i] = 0;
        for (j = 0; j < n; j++) {
            out[i] += in[j]*cos(i*pn*(j+0.5));
        }
    }
}